/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Arrays;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import com.koutra.dist.proc.faucet.ReaderFaucet;
import com.koutra.dist.proc.faucet.StreamFaucet;
import com.koutra.dist.proc.pipeline.ReaderDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.ReaderToWriterPipelineItem;
import com.koutra.dist.proc.pipeline.ReaderToWriterPipelineItemTemplate;
import com.koutra.dist.proc.pipeline.SplitReaderPipelineItem;
import com.koutra.dist.proc.pipeline.SplitReaderPipelineItemTemplate;
import com.koutra.dist.proc.pipeline.SplitStreamPipelineItem;
import com.koutra.dist.proc.pipeline.SplitStreamPipelineItemTemplate;
import com.koutra.dist.proc.pipeline.StreamDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.StreamToStreamPipelineItem;
import com.koutra.dist.proc.pipeline.StreamToStreamPipelineItemTemplate;
import com.koutra.dist.proc.sink.StreamMuxSink;
import com.koutra.dist.proc.sink.StreamSink;
import com.koutra.dist.proc.sink.WriterMuxSink;
import com.koutra.dist.proc.sink.WriterSink;
import com.koutra.dist.proc.util.Util;

import junit.framework.TestCase;

public class SplitTest extends TestCase {
	public void testSplitStream() throws IOException, InterruptedException, ExecutionException {
		// Set up the stream objects
		StreamFaucet faucet = new StreamFaucet("1", "test/input.txt");
		SplitStreamPipelineItem splitter = new SplitStreamPipelineItem("2", 2);
		StreamToStreamPipelineItem transformerA = new StreamToStreamPipelineItem("1.1") {
			@Override
			public byte[] transformBuffer(byte[] buffer, int off, int len) {
				if (len == -1)
					return new byte[0];
				return Arrays.copyOfRange(buffer, off, off + len);
			}
		};
		StreamToStreamPipelineItem transformerB = new StreamToStreamPipelineItem("2.1") {
			@Override
			public byte[] transformBuffer(byte[] buffer, int off, int len) {
				if (len == -1)
					return new byte[0];
				return Arrays.copyOfRange(buffer, off, off + len);
			}
		};
		StreamSink sinkA = new StreamSink("1.2", "test/outputA.txt");
		StreamSink sinkB = new StreamSink("2.2", "test/outputB.txt");
		
		// Hook everything up
		splitter.hookupFaucet(faucet);
		splitter.hookupSink(0, transformerA);
		splitter.hookupSink(1, transformerB);
		transformerA.hookupSink(sinkA);
		transformerB.hookupSink(sinkB);
		
		// Start the ball rolling
		Future<?> futureA = sinkA.consumeAsynchronously();
		Future<?> futureB = sinkB.consumeAsynchronously();
		
		futureA.get();
		futureB.get();
		
		// Shut things down
		sinkA.dispose();
		sinkB.dispose();
		
		// Verify the result
		InputStream inputFileIS = new FileInputStream("test/input.txt");
		InputStream outputFileIS = new FileInputStream("test/outputA.txt");
		Util.testStreamEquivalence(inputFileIS, outputFileIS);
		inputFileIS.close();
		outputFileIS.close();
		
		// Verify the result
		inputFileIS = new FileInputStream("test/input.txt");
		outputFileIS = new FileInputStream("test/outputB.txt");
		Util.testStreamEquivalence(inputFileIS, outputFileIS);
		inputFileIS.close();
		outputFileIS.close();

		File fileToDelete = new File("test/outputA.txt");
		fileToDelete.delete();
		
		fileToDelete = new File("test/outputB.txt");
		fileToDelete.delete();
	}
	
	public static class IdentityStreamToStreamPipelineItemTemplate
			extends StreamToStreamPipelineItemTemplate {
		@Override
		public byte[] transformBuffer(byte[] buffer, int off, int len) {
			if (len == -1)
				return new byte[0];
			return Arrays.copyOfRange(buffer, off, off + len);
		}
	}
	
	public void testSplitStreamTemplate() throws IOException, InterruptedException, ExecutionException {
		// Set up the stream objects
		StreamFaucet faucet = new StreamFaucet("1", "test/input.txt");
		SplitStreamPipelineItemTemplate splitter = new SplitStreamPipelineItemTemplate(2);
		StreamDemuxPipelineItem demux = new StreamDemuxPipelineItem("2", splitter) {
			@Override
			public boolean shouldChangePipeline() {
				return false;
			}
			@Override
			public byte[] transformBuffer(byte[] buffer, int off, int len) {
				if (len == -1)
					return new byte[0];
				
				return Arrays.copyOfRange(buffer, off, len);
			}
		};
		IdentityStreamToStreamPipelineItemTemplate transformerA =
			new IdentityStreamToStreamPipelineItemTemplate();
		IdentityStreamToStreamPipelineItemTemplate transformerB =
			new IdentityStreamToStreamPipelineItemTemplate();
		StreamMuxSink sinkA = new StreamMuxSink("1.2", "test/outputA.txt");
		StreamMuxSink sinkB = new StreamMuxSink("2.2", "test/outputB.txt");
		
		// Hook everything up
		demux.hookupFaucet(faucet);
		splitter.hookupFaucet(demux);
		splitter.hookupSink(0, transformerA);
		splitter.hookupSink(1, transformerB);
		transformerA.hookupSink(sinkA);
		transformerB.hookupSink(sinkB);
		
		// Start the ball rolling
		Future<?> futureA = sinkA.consumeAsynchronously();
		Future<?> futureB = sinkB.consumeAsynchronously();
		
		futureA.get();
		futureB.get();
		
		// Shut things down
		sinkA.dispose();
		sinkB.dispose();
		
		// Verify the result
		InputStream inputFileIS = new FileInputStream("test/input.txt");
		InputStream outputFileIS = new FileInputStream("test/outputA.txt");
		Util.testStreamEquivalence(inputFileIS, outputFileIS);
		inputFileIS.close();
		outputFileIS.close();
		
		// Verify the result
		inputFileIS = new FileInputStream("test/input.txt");
		outputFileIS = new FileInputStream("test/outputB.txt");
		Util.testStreamEquivalence(inputFileIS, outputFileIS);
		inputFileIS.close();
		outputFileIS.close();

		File fileToDelete = new File("test/outputA.txt");
		fileToDelete.delete();
		
		fileToDelete = new File("test/outputB.txt");
		fileToDelete.delete();
	}
	
	public void testSplitReader() throws IOException, InterruptedException, ExecutionException {
		// Set up the stream objects
		ReaderFaucet faucet = new ReaderFaucet("1", "test/greek_input.txt", "UTF-8");
		SplitReaderPipelineItem splitter = new SplitReaderPipelineItem("2", 2);
		ReaderToWriterPipelineItem transformerA = new ReaderToWriterPipelineItem("1.1") {
			@Override
			public char[] transformBuffer(char[] buffer, int off, int len) {
				if (len == -1)
					return new char[0];
				return Arrays.copyOfRange(buffer, off, off + len);
			}
		};
		ReaderToWriterPipelineItem transformerB = new ReaderToWriterPipelineItem("2.1") {
			@Override
			public char[] transformBuffer(char[] buffer, int off, int len) {
				if (len == -1)
					return new char[0];
				return Arrays.copyOfRange(buffer, off, off + len);
			}
		};
		WriterSink sinkA = new WriterSink("1.2", "test/greek_outputA.txt", "Cp1253");
		WriterSink sinkB = new WriterSink("2.2", "test/greek_outputB.txt", "Cp1253");
		
		// Hook everything up
		splitter.hookupFaucet(faucet);
		splitter.hookupSink(0, transformerA);
		splitter.hookupSink(1, transformerB);
		transformerA.hookupSink(sinkA);
		transformerB.hookupSink(sinkB);
		
		// Start the ball rolling
		Future<?> futureA = sinkA.consumeAsynchronously();
		Future<?> futureB = sinkB.consumeAsynchronously();
		
		futureA.get();
		futureB.get();
		
		// Shut things down
		sinkA.dispose();
		sinkB.dispose();
		
		// Verify the result
		Reader inputFileReader = new InputStreamReader(
				new FileInputStream("test/greek_input.txt"), "UTF-8");
		Reader outputFileReader = new InputStreamReader(
				new FileInputStream("test/greek_outputA.txt"), "Cp1253");
		Util.testReaderEquivalence(inputFileReader, outputFileReader);
		inputFileReader.close();
		outputFileReader.close();
		
		// Verify the result
		inputFileReader = new InputStreamReader(
				new FileInputStream("test/greek_input.txt"), "UTF-8");
		outputFileReader = new InputStreamReader(
				new FileInputStream("test/greek_outputB.txt"), "Cp1253");
		Util.testReaderEquivalence(inputFileReader, outputFileReader);
		inputFileReader.close();
		outputFileReader.close();

		File fileToDelete = new File("test/greek_outputA.txt");
		fileToDelete.delete();
		
		fileToDelete = new File("test/greek_outputB.txt");
		fileToDelete.delete();
	}
	
	public static class IdentityReaderToWriterPipelineItemTemplate
		extends ReaderToWriterPipelineItemTemplate {
		@Override
		public char[] transformBuffer(char[] buffer, int off, int len) {
			if (len == -1)
				return new char[0];
			return Arrays.copyOfRange(buffer, off, off + len);
		}
	}

	public void testSplitReaderTemplate() throws IOException, InterruptedException, ExecutionException {
		// Set up the stream objects
		ReaderFaucet faucet = new ReaderFaucet("1", "test/greek_input.txt", "UTF-8");
		SplitReaderPipelineItemTemplate splitter = new SplitReaderPipelineItemTemplate(2);
		ReaderDemuxPipelineItem demux = new ReaderDemuxPipelineItem("2", splitter) {
			@Override
			public boolean shouldChangePipeline() {
				return false;
			}
			@Override
			public char[] transformBuffer(char[] buffer, int off, int len) {
				if (len == -1)
					return new char[0];
				
				return Arrays.copyOfRange(buffer, off, len);
			}
		};
		IdentityReaderToWriterPipelineItemTemplate transformerA =
			new IdentityReaderToWriterPipelineItemTemplate();
		IdentityReaderToWriterPipelineItemTemplate transformerB =
			new IdentityReaderToWriterPipelineItemTemplate();
		WriterMuxSink sinkA = new WriterMuxSink("1.2", "test/greek_outputA.txt", "Cp1253");
		WriterMuxSink sinkB = new WriterMuxSink("2.2", "test/greek_outputB.txt", "Cp1253");
		
		// Hook everything up
		demux.hookupFaucet(faucet);
		splitter.hookupFaucet(demux);
		splitter.hookupSink(0, transformerA);
		splitter.hookupSink(1, transformerB);
		transformerA.hookupSink(sinkA);
		transformerB.hookupSink(sinkB);
		transformerB.hookupSink(sinkB);
		
		// Start the ball rolling
		Future<?> futureA = sinkA.consumeAsynchronously();
		Future<?> futureB = sinkB.consumeAsynchronously();
		
		futureA.get();
		futureB.get();
		
		// Shut things down
		sinkA.dispose();
		sinkB.dispose();
		
		// Verify the result
		Reader inputFileReader = new InputStreamReader(
				new FileInputStream("test/greek_input.txt"), "UTF-8");
		Reader outputFileReader = new InputStreamReader(
				new FileInputStream("test/greek_outputA.txt"), "Cp1253");
		Util.testReaderEquivalence(inputFileReader, outputFileReader);
		inputFileReader.close();
		outputFileReader.close();
		
		// Verify the result
		inputFileReader = new InputStreamReader(
				new FileInputStream("test/greek_input.txt"), "UTF-8");
		outputFileReader = new InputStreamReader(
				new FileInputStream("test/greek_outputB.txt"), "Cp1253");
		Util.testReaderEquivalence(inputFileReader, outputFileReader);
		inputFileReader.close();
		outputFileReader.close();

		File fileToDelete = new File("test/greek_outputA.txt");
		fileToDelete.delete();
		
		fileToDelete = new File("test/greek_outputB.txt");
		fileToDelete.delete();
	}

}
