/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.cluster;

import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringWriter;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.stream.StreamResult;

import org.apache.log4j.Logger;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLFilterImpl;

import com.koutra.dist.proc.util.ByteArrayUtil;
import com.koutra.dist.proc.util.DumpingOutputStream;

/**
 * Used in conjunction with a <code>PipedInputXMLStream</code>, this class is an XMLFilter
 * that can write the SAX event stream to a socket, thereby allowing pipelines
 * to perform part of their computation on a remote JVM.
 * 
 * @author Pafsanias Ftakas
 */
public class PipedOutputXMLFilter extends XMLFilterImpl {
	private static final Logger logger = Logger.getLogger(PipedOutputXMLFilter.class);
	
	protected ServerSocket socket;
	protected Socket clientSocket;
	protected InetAddress otherEndHostIP;
	protected int otherEndPort;
	
	public PipedOutputXMLFilter() throws IOException {
		socket = new ServerSocket(0);
		
		if (logger.isDebugEnabled())
			logger.debug("Started server socket at port: " + socket.getLocalPort());
		
		clientSocket = null;
	}
	
	public byte[] getServerHostIP() {
		return socket.getInetAddress().getAddress();
	}
	
	public int getServerPort() {
		return socket.getLocalPort();
	}
	
	public byte[] getOtherEndHostIP() {
		return otherEndHostIP.getAddress();
	}
	
	public int getOtherEndPort() {
		return otherEndPort;
	}
	
	/**
	 * Connect the server socket to the client. This method potentially blocks until the
	 * corresponding <code>PipedInputXMLFilter</code> connects to us. After connecting with the
	 * client, this class receives the remote end information from the other side.
	 * This information is "don't care" for a task boundary that sits at the end of
	 * a demux/mux pipeline, but is very important for a task boundary that sits at
	 * the begining of the demux/mux pipeline, as it contains the server socket information
	 * for the server socket sitting at the task boundary at the end.
	 * @throws IOException iff something goes wrong during the connection to the client
	 * and the receiving of the remote information.
	 */
	public void connect() throws IOException {
		clientSocket = socket.accept();
		
		if (logger.isDebugEnabled())
			logger.debug("Accepted socket: " +
					clientSocket.getLocalAddress() + ":" + clientSocket.getLocalPort());

		Reader reader = new InputStreamReader(clientSocket.getInputStream());
		StringWriter writer = new StringWriter();
		int count;
		char[] buffer = new char[8 * 1024];
		while ((count = reader.read(buffer)) != -1) {
			writer.write(buffer, 0, count);
			if (writer.toString().indexOf('\n') != -1)
				break;
		}
		// DO NOT close the reader here. It will close the damn socket!
		
		if (logger.isTraceEnabled())
			logger.trace("Read '" + writer.toString().trim() + "' from the socket.");
		
		String readInfo = writer.toString().substring(0, writer.toString().indexOf('\n'));
		String remoteIPString = readInfo.substring(0, readInfo.indexOf('|'));
		String remotePortString = readInfo.substring(
				readInfo.indexOf('|') + 1, readInfo.length() - 1);
		
		otherEndHostIP = InetAddress.getByAddress(
				ByteArrayUtil.decodeFromHex(remoteIPString));
		otherEndPort = Integer.parseInt(remotePortString);
	}
	
	public void drivePipeline(InputSource inputSource) throws IOException, TransformerException {
		OutputStream outputStream = new DumpingOutputStream(new BufferedOutputStream(
				clientSocket.getOutputStream()));
		
		// Set up the output stream
	    StreamResult result = new StreamResult(outputStream);
	    
	    SAXTransformerFactory stf = (SAXTransformerFactory) TransformerFactory.newInstance();
	    Transformer transformer = stf.newTransformer();
	    SAXSource transformSource = new SAXSource((XMLReader) this, inputSource);
	    transformer.transform(transformSource, result);
	    
	    outputStream.close();
	}
	
	@Override
	public void parse(InputSource input) throws SAXException, IOException {
		super.parse(input);
	}
	
	@Override
	public void startDocument() throws SAXException {
		// TODO Auto-generated method stub
		super.startDocument();
	}
	
	@Override
	public void startElement(String uri, String localName, String name,
			Attributes atts) throws SAXException {
		// TODO Auto-generated method stub
		super.startElement(uri, localName, name, atts);
	}

}
