/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.faucet;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.List;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.ISplitPipelineItem;
import com.koutra.dist.proc.model.XformationException;

/**
 * A faucet implementation that uses the java.io.InputStream interface to generate
 * content for the transformation. When using the constructor that supplies a
 * java.io.InputStream, the caller has the responsibility of closing the stream once
 * the transformation has completed.
 * 
 * @author Pafsanias Ftakas
 */
public class StreamFaucet implements IFaucet {
	private static final Logger logger = Logger.getLogger(StreamFaucet.class);
	
	protected enum Type {
		File,
		InputStream;
	};
	
	protected String id;
	protected ISink sink;
	protected boolean hookedUp;
	protected Type type;
	protected InputStream inputStream;
	protected String path;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public StreamFaucet() { }
	
	public StreamFaucet(String id, InputStream is) {
		this.id = id;
		this.sink = null;
		this.hookedUp = false;
		this.type = Type.InputStream;
		this.inputStream = is;
		this.path = null;
	}

	public StreamFaucet(String id, String path) {
		this.id = id;
		this.sink = null;
		this.hookedUp = false;
		this.type = Type.File;
		this.inputStream = null;
		this.path = path;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public ISink getSink() {
		return sink;
	}

	@Override
	public boolean supportsOutput(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
			return true;
		case CharStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public void hookupSink(ISink sink) {
		if (!sink.supportsInput(ContentType.ByteStream))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' must support the ByteStream content type.");
		if (sink instanceof IMuxSink)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an IMuxSink instance. This is not allowed.");
		if (sink instanceof ISinkTemplate)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an ISinkTemplate instance. This is not allowed.");
		if (hookedUp && this.sink == sink) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUp)
			throw new XformationException("Trying to hook up an already hooked up faucet.");
		
		this.sink = sink;
		this.hookedUp = true;
		
		this.sink.hookupFaucet(this);
	}
	
	@Override
	public Object getSource(ContentType contentType) {
		switch (contentType) {
		case CharStream:
		case XML:
		case ResultSet:
			return null;
		}
		
		if (inputStream != null)
			return inputStream;

		switch (type) {
		case File:
			try {
				FileInputStream i = new FileInputStream(path);
				inputStream = new BufferedInputStream(i);
				
				if (logger.isTraceEnabled())
					logger.trace("Faucet has created reader " + inputStream +
							" wrapping reader " + i + " reading from path " + path);

			} catch (FileNotFoundException e) {
				throw new XformationException("Unable to create input stream", e);
			}
			break;
		case InputStream:
			break;
		}
		
		return inputStream;
	}
	
	@Override
	public void dispose() {
		switch (type) {
		case File:
			try {
				inputStream.close();
			} catch (IOException e) {
				throw new XformationException("Unable to close input stream", e);
			}
			break;
		case InputStream:
			// Caller owns the input stream. It is their responsibility to close it.
			break;
		}
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		String sinkClassName = in.readUTF();
		Class<?> sinkClass;
		try {
			sinkClass = Class.forName(sinkClassName);
		} catch (ClassNotFoundException e) {
			throw new InstantiationException("Unable to load class: " + sinkClassName);
		}
		this.sink = (ISink) sinkClass.newInstance();
		this.sink.readFrom(in);
		this.hookedUp = true;
		this.sink.hookupFaucet(this);
		this.type = Type.File;
		this.inputStream = null;
		this.path = in.readUTF();
	}

	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		if (type != Type.File)
			throw new XformationException(
					"Only StreamFaucets of type Type.File can be streamed.");
		
		out.writeUTF(id);
		out.writeUTF(sink.getClass().getCanonicalName());
		sink.writeTo(out);
		out.writeUTF(path);
	}

	@Override
	public List<ISink> getTerminalSinks() {
		if (sink instanceof IFaucet)
			return ((IFaucet) sink).getTerminalSinks();
		else if (sink instanceof ISplitPipelineItem)
			return ((ISplitPipelineItem) sink).getTerminalSinks();
		else
			return Collections.singletonList(sink);
	}

}
