/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.pipeline;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Future;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.cluster.ClusterRunnable;
import com.koutra.dist.proc.cluster.IPipedInput;
import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.FaucetAdapter;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.IPipelineItemTemplate;
import com.koutra.dist.proc.model.ISimplePipelineItem;
import com.koutra.dist.proc.model.ISimplePipelineItemTemplate;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.ISplitPipelineItem;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.ExecutorsHelper;

public abstract class AbstractTaskBoundaryTemplate implements ISimplePipelineItemTemplate {
	private static final Logger logger =
		Logger.getLogger(AbstractTaskBoundaryTemplate.class);
	
	protected String id;
	protected boolean isClone;
	protected boolean hookedUpFaucet;
	protected boolean hookedUpSink;
	protected ISink sink;
	protected IFaucet faucet;
	protected boolean isRemoteTaskSide;
	protected byte[] remoteServerIP;
	protected int remoteServerPort;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public AbstractTaskBoundaryTemplate() { }
	
	public AbstractTaskBoundaryTemplate(String id) {
		this.id = id;
		this.isClone = false;
		this.hookedUpFaucet = false;
		this.hookedUpSink = false;
		this.sink = null;
		this.faucet = null;
		this.isRemoteTaskSide = false;
		this.remoteServerIP = null;
		this.remoteServerPort = 0;
	}

	@Override
	public boolean isClone() {
		return isClone;
	}
	
	protected abstract ContentType getSupportedFaucetContentType();

	@Override
	public void createClone(IPipelineItem pipelineItem) {
		try {
			final AbstractTaskBoundaryTemplate clone = getClass().newInstance();
			clone.id = id;
			clone.isClone = true;
			clone.hookedUpFaucet = false;
			clone.hookedUpSink = false;
			clone.sink = null;
			clone.faucet = null;
			clone.isRemoteTaskSide = isRemoteTaskSide;
			clone.remoteServerIP = null;
			clone.remoteServerPort = 0;
			
			if (pipelineItem instanceof ISimplePipelineItem)
				((ISimplePipelineItem) pipelineItem).hookupSink(clone);
			else if (pipelineItem instanceof FaucetAdapter)
				((FaucetAdapter) pipelineItem).hookupSink(clone);
			
			if (sink instanceof ISinkTemplate) {
				((ISinkTemplate) sink).createClone(clone);
			} else {
				sink.hookupFaucet(clone);
			}
			
			if (!isRemoteTaskSide && !(sink instanceof IMuxSink)) {
				/*
				if (logger.isTraceEnabled()) {
					String position = (sink instanceof IMuxSink ? "end" : "start") +
							" of pipeline";
					String location = isRemoteTaskSide ? "remote" : "local";
					logger.trace("this: " + this + ", outputFilter: " + outputFilter);
					try {
						throw new Exception(position + ", " + location);
					} catch (Exception e) {
						logger.trace("Calling clone for: ", e);
					}
				}
				*/
				
				clone.getSource(getSupportedFaucetContentType());
				
				ExecutorsHelper.getInstance().executeInCluster(new ClusterRunnable(clone));
			}
		} catch (Exception e) {
			throw new XformationException("Unable to create clone", e);
		}
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public ISink getSink() {
		return sink;
	}

	@Override
	public IFaucet getFaucet() {
		return faucet;
	}

	protected abstract void checkFaucetValidity(IFaucet faucet);

	@Override
	public void hookupFaucet(IFaucet faucet) {
		checkFaucetValidity(faucet);
		if (hookedUpFaucet && this.faucet == faucet) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUpFaucet)
			throw new XformationException("Trying to hook up an already hooked up faucet.");
		
		this.faucet = faucet;
		this.hookedUpFaucet = true;
		
		this.faucet.hookupSink(this);
	}

	protected abstract void checkSinkValidity(ISink sink);

	@Override
	public void hookupSink(ISink sink) {
		checkSinkValidity(sink);
		if (hookedUpSink && this.sink == sink) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUpSink)
			throw new XformationException("Trying to hook up an already hooked up sink.");
		
		this.sink = sink;
		this.hookedUpSink = true;
		
		this.sink.hookupFaucet(this);
	}

	@Override
	public void registerSource(Object source) {
		// only used in mux/demux cases.
	}
	
	protected abstract byte[] getOtherEndHostIP();
	protected abstract int getOtherEndPort();
	
	@Override
	public void handleMuxSwitch() {
		if (sink instanceof IPipelineItem) {
			if (!isRemoteTaskSide) {
				// Go down the sink pipeline until you read the task boundary. Set
				// the host IP and port number in that site to the ones retrieved from
				// our piped stream pair. This will allow the generation of the appropriate
				// input stream that can connect to the right server socket on the cluster
				// side, so that the IMuxSink instance can consume the cluster pipeline output.
				IPipelineItemTemplate template = (IPipelineItemTemplate) sink;
				while (!(template instanceof AbstractTaskBoundaryTemplate)) {
					if (template instanceof ISimplePipelineItemTemplate)
						template = (IPipelineItemTemplate)
								((ISimplePipelineItemTemplate) template).getSink();
					else
						throw new XformationException("Unexpected template type: " +
								template.getClass().getCanonicalName());
				}
				AbstractTaskBoundaryTemplate boundaryTemplate =
					(AbstractTaskBoundaryTemplate) template;
				
				boundaryTemplate.remoteServerIP = getOtherEndHostIP();
				boundaryTemplate.remoteServerPort = getOtherEndPort();
			}

			((IPipelineItem) sink).handleMuxSwitch();
		}
	}

	@Override
	public void endMux() {
		if (sink instanceof IPipelineItem)
			((IPipelineItemTemplate) sink).endMux();
		else if (!isRemoteTaskSide)
			sink.registerSource(null);
	}

	@Override
	public void consume() {
		throw new XformationException("Calling ISink.consume() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}

	@Override
	public Future<?> consumeAsynchronously() {
		throw new XformationException("Calling ISink.consumeAsynchronously() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}
	
	protected abstract byte[] getServerHostIP();
	protected abstract int getServerPort();
	protected abstract void localStartBoundaryImpl(Object inputSource);
	protected abstract void remoteEndBoundaryImpl(Object inputSource);
	
	protected Object initialInputSource() {
		return null;
	}
	
	@Override
	public Object consume(ISink previousSink) {
		Object inputSource = initialInputSource();
		if (faucet instanceof IPipelineItem)
			inputSource = ((IPipelineItem) faucet).consume(this);
	    
	    if (this instanceof ISinkTemplate) {
	    	if (!((ISinkTemplate) this).isClone()) {
	    		return null;
	    	}
	    }

		if (!hookedUpSink)
			throw new XformationException("Pipeline item has not been set up " +
					"correctly: sink has not been set");
		
		/*
		 * Useful for debugging purposes, but confusing for normal output
		if (logger.isTraceEnabled()) {
			String position = (sink instanceof IMuxSink ? "end" : "start") +
					" of pipeline";
			String location = isRemoteTaskSide ? "remote" : "local";
			logger.trace("this: " + this + ", outputFilter: " + outputFilter);
			try {
				throw new Exception(position + ", " + location);
			} catch (Exception e) {
				logger.trace("Calling consume for: ", e);
			}
		}
		*/
		
		if (!(sink instanceof IMuxSink)) {
			// We are at the start of the pipeline
			if (isRemoteTaskSide) {
				// Go down the sink pipeline until you read the task boundary. Retrieve
				// the host IP and port number from that site and communicate it back
				// to the original pipeline for usage in the generation of the final
				// piped input stream.
				IPipelineItemTemplate template = (IPipelineItemTemplate) sink;
				while (!(template instanceof AbstractTaskBoundaryTemplate)) {
					if (template instanceof ISimplePipelineItemTemplate)
						template = (IPipelineItemTemplate)
								((ISimplePipelineItemTemplate) template).getSink();
					else
						throw new XformationException("Unexpected template type: " +
								template.getClass().getCanonicalName());
				}
				AbstractTaskBoundaryTemplate boundaryTemplate =
					(AbstractTaskBoundaryTemplate) template;
				boundaryTemplate.getSource(boundaryTemplate.getSupportedFaucetContentType());
				byte[] otherEndHostIP = boundaryTemplate.getServerHostIP();
				int otherEndPort = boundaryTemplate.getServerPort();
				
				if (logger.isTraceEnabled()) {
					try {
						logger.trace("Retrieved from the end of the pipeline boundary " +
								"template IP: " +
								InetAddress.getByAddress(otherEndHostIP).toString() +
								":" + otherEndPort);
					} catch (UnknownHostException e1) {
						logger.error("Unable to find host to connect to in the " +
								"cluster pipeline.", e1);
					}
				}

				IPipedInput pis = (IPipedInput) getSource(getSupportedFaucetContentType());
				try {
					pis.communicateOtherEndDetails(otherEndHostIP, otherEndPort);
				} catch (IOException e) {
					throw new XformationException("Unable to communicate other end " +
							"details to the original pipeline.", e);
				}
				
				return inputSource;
			} else { // !isRemoteTaskSide
				// Now consume the faucet input to terminate the threads started by our sinks.
				final Object inputSourceToUse = inputSource;
				
				ExecutorsHelper.getInstance().executeInProc(new Runnable() {
					@Override
					public void run() {
						if (logger.isTraceEnabled())
							logger.trace("Driving pipeline up to the starting boundary");
						
						localStartBoundaryImpl(inputSourceToUse);
						
						logger.debug("Executable for the XML task boundary " +
								"(start of pipeline) will now exit.");
					}
				});
				
				return inputSource;
			}
		} else { // sink instanceof IMuxSink
			// We are at the end of the pipeline, just before the mux sink
			if (isRemoteTaskSide) {
				remoteEndBoundaryImpl(inputSource);

				return inputSource;
			} else { // !isRemoteTaskSide
				IPipedInput pis = (IPipedInput) getSource(getSupportedFaucetContentType());
				try {
					// We do not care about this pair, as they have exchanged details
					// during the stream (de)serialization to the cluster. Just send
					// some valid nonsense.
					pis.communicateOtherEndDetails(
							InetAddress.getLocalHost().getAddress(), 0);
				} catch (IOException e) {
					throw new XformationException("Unable to communicate other end " +
							"details to the original pipeline.", e);
				}
				
				return inputSource;
			}
		}
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		this.isClone = in.readBoolean();
		String sinkClassName = in.readUTF();
		Class<?> sinkClass;
		try {
			sinkClass = Class.forName(sinkClassName);
		} catch (ClassNotFoundException e) {
			throw new InstantiationException("Unable to load class: " + sinkClassName);
		}
		this.sink = (ISink) sinkClass.newInstance();
		this.sink.readFrom(in);
		this.hookedUpSink = true;
		this.sink.hookupFaucet(this);
		this.faucet = null;
		this.hookedUpFaucet = false;
		this.isRemoteTaskSide = true;
		this.remoteServerIP = null;
		this.remoteServerPort = 0;
	}
	
	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		out.writeUTF(id);
		out.writeBoolean(isClone);
		out.writeUTF(sink.getClass().getCanonicalName());
		sink.writeTo(out);
	}

	@Override
	public List<ISink> getTerminalSinks() {
		if (sink instanceof IFaucet)
			return ((IFaucet) sink).getTerminalSinks();
		else if (sink instanceof ISplitPipelineItem)
			return ((ISplitPipelineItem) sink).getTerminalSinks();
		else
			return Collections.singletonList(sink);
	}

}
