/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.pipeline;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicReference;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.sax.SAXTransformerFactory;

import org.apache.log4j.Logger;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISimplePipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.ISplitPipelineItem;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.ExecutorsHelper;
import com.koutra.dist.proc.xml.IStreamDemuxSAXCallback;
import com.koutra.dist.proc.xml.SAXEventQueueReader;
import com.koutra.dist.proc.xml.StreamSAXEventReader;
import com.koutra.dist.proc.xml.XMLDemuxContentHandler;

public abstract class StreamToXMLDemuxPipelineItem implements ISimplePipelineItem,
		IDemuxFaucet, IStreamDemuxSAXCallback {
	private static final Logger logger = Logger.getLogger(
			StreamToXMLDemuxPipelineItem.class);

	protected String id;
	protected boolean hookedUpFaucet;
	protected boolean hookedUpSink;
	protected AtomicReference<ISink> sink;
	protected ISinkTemplate pipelineTemplate;
	protected IFaucet faucet;
	protected XMLReader reader;
	protected XMLDemuxContentHandler demuxContentHandler;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public StreamToXMLDemuxPipelineItem() { }
	
	public StreamToXMLDemuxPipelineItem(String id, ISinkTemplate pipelineTemplate) {
		if (!pipelineTemplate.supportsInput(ContentType.XML))
			throw new IllegalArgumentException("SinkTemplate '" +
					pipelineTemplate.getId() +
					"' must support the XML content type.");
		this.id = id;
		this.hookedUpFaucet = false;
		this.hookedUpSink = false;
		this.sink = new AtomicReference<ISink>(null);
		this.faucet = null;
		this.pipelineTemplate = pipelineTemplate;
		this.reader = null;
		this.demuxContentHandler = null;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public ISink getSink() {
		return sink.get();
	}

	@Override
	public IFaucet getFaucet() {
		return faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
			return true;
		case CharStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public boolean supportsOutput(ContentType contentType) {
		switch (contentType) {
		case XML:
			return true;
		case ByteStream:
		case CharStream:
		case ResultSet:
		default:
			return false;
		}
	}

	protected void checkFaucetValidity(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.ByteStream))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the ByteStream content type.");
		if (faucet instanceof IDemuxFaucet)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IDemuxFaucet instance. This is not allowed.");
		if (faucet instanceof IFaucetTemplate)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IFaucetTemplate instance. This is not allowed.");
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		checkFaucetValidity(faucet);
		if (hookedUpFaucet && this.faucet == faucet) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUpFaucet)
			throw new XformationException("Trying to hook up an already hooked up sink.");
		
		this.faucet = faucet;
		this.hookedUpFaucet = true;
		
		this.faucet.hookupSink(this);
	}

	protected void checkSinkValidity(ISink sink) {
		if (!sink.supportsInput(ContentType.XML))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' must support the XML content type.");
		if (!(sink instanceof IMuxSink) && !(sink instanceof ISinkTemplate))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is neither an IMuxSink or an ISinkTemplate instance. " +
					"This is not allowed.");
	}
	
	@Override
	public void hookupSink(ISink sink) {
		checkSinkValidity(sink);
		if (hookedUpSink && this.sink.get() == sink) {
			// Effectively a NO-OP.
			return;
		}
		
		this.sink.set(sink);
		this.hookedUpSink = true;
		
		this.sink.get().hookupFaucet(this);
	}
	
	@Override
	public Object getSource(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
		case CharStream:
		case ResultSet:
			throw new XformationException("Content type: " + contentType +
					" is not supported.");
		}

		try {
			reader = new SAXEventQueueReader(demuxContentHandler.getNextEventList());
		} catch (InterruptedException ie) {
			throw new XformationException("Unable to create reader.", ie);
		}
		
		return reader;
	}
	
	@Override
	public void registerSource(Object source) {
		// Only used in mux/demux cases.
	}
	
	@Override
	public void handleMuxSwitch() {
		// Only used in mux/demux cases.
	}
	
	@Override
	public void endMux() {
		// only used in mux/demux cases.
	}
	
	protected void createPipeline() {
		pipelineTemplate.createClone(this);
	}
	
	public void switchPipeline() {
		sink.set(null);
		hookedUpSink = false;
		
		createPipeline();
		
		if (sink.get() instanceof IPipelineItem)
			((IPipelineItem) sink.get()).handleMuxSwitch();
		
		if (logger.isTraceEnabled())
			logger.trace("Dumping pipeline:\n" + dumpPipeline());
	}
	
	@Override
	public String dumpPipeline() {
		return sink.get().dumpPipeline() + "\n" + getClass().getName() + ": " + reader;
	}
	
	public void signalTermination() {
		// Send a null down to the mux sink so that it knows that the ball
		// has stopped rolling.
		if (pipelineTemplate instanceof IPipelineItem)
			((IPipelineItem) pipelineTemplate).endMux();
		else
			pipelineTemplate.registerSource(null);
	}
	
	@Override
	public void dispose() {
		ISink currentSink = sink.get();
		if (currentSink == null)
			return;
		if (currentSink instanceof ISinkTemplate)
			if (((ISinkTemplate) currentSink).isClone())
				return;
		
		faucet.dispose();
	}

	@Override
	public void consume() {
		throw new XformationException("Calling ISink.consume() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}
	
	@Override
	public Future<?> consumeAsynchronously() {
		throw new XformationException("Calling ISink.consumeAsynchronously() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}

	@Override
	public Object consume(ISink previousSink) {
		if (!hookedUpFaucet)
			throw new XformationException("Pipeline item has not been set up " +
					"correctly: faucet has not been set");
		
		if (previousSink instanceof ISinkTemplate) {
			// If the sink is a clone we must stop the consumption chain here, as the
			// rest of the chain has already been notified the first time around.
			if (((ISinkTemplate) previousSink).isClone()) {
				return new InputSource();
			}
		}
		
		demuxContentHandler = new XMLDemuxContentHandler();
		
		createPipeline();

		// Set up the output stream
	    final SAXResult result = new SAXResult(demuxContentHandler);
	    
		final InputStream inputStream = (InputStream) faucet.getSource(ContentType.ByteStream);

		ExecutorsHelper.getInstance().executeInProc(new Runnable() {
			@Override
			public void run() {
			    // Set up the transformer to process the SAX events generated
			    // by the last filter in the chain
			    try {
			    	XMLReader reader = new StreamSAXEventReader(
			    			inputStream, StreamToXMLDemuxPipelineItem.this,
			    			StreamToXMLDemuxPipelineItem.this);

			    	if (logger.isTraceEnabled())
						logger.trace("Creating SAX event reader " + reader +
								" mapping input stream: " + inputStream);

				    SAXSource transformSource = new SAXSource(
			    			reader, new InputSource());
				    SAXTransformerFactory stf = (SAXTransformerFactory)
				    	TransformerFactory.newInstance();
				    Transformer transformer = stf.newTransformer();
				    transformer.transform(transformSource, result);
			    } catch (Exception e) {
			    	throw new XformationException("Unable to set up transform", e);
			    }
			}
		});
	    
	    return null;
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		this.hookedUpFaucet = false;
		this.hookedUpSink = false;
		this.sink = new AtomicReference<ISink>(null);
		this.faucet = null;
		this.reader = null;
		this.demuxContentHandler = null;

		String sinkClassName = in.readUTF();
		Class<?> sinkClass;
		try {
			sinkClass = Class.forName(sinkClassName);
		} catch (ClassNotFoundException e) {
			throw new InstantiationException("Unable to load class: " + sinkClassName);
		}
		this.pipelineTemplate = (ISinkTemplate) sinkClass.newInstance();
		this.pipelineTemplate.readFrom(in);
		this.pipelineTemplate.hookupFaucet(this);
	}
	
	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		out.writeUTF(id);
		out.writeUTF(pipelineTemplate.getClass().getCanonicalName());
		pipelineTemplate.writeTo(out);
	}

	@Override
	public List<ISink> getTerminalSinks() {
		if (pipelineTemplate instanceof IFaucet)
			return ((IFaucet) pipelineTemplate).getTerminalSinks();
		else if (pipelineTemplate instanceof ISplitPipelineItem)
			return ((ISplitPipelineItem) pipelineTemplate).getTerminalSinks();
		else {
			List<ISink> retVal = new ArrayList<ISink>();
			retVal.add(pipelineTemplate);
			return retVal;
		}
	}

}
