/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.sink;

import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.concurrent.BlockingDeque;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.sax.SAXSource;

import org.apache.fop.apps.FOUserAgent;
import org.apache.fop.apps.Fop;
import org.apache.fop.apps.FopFactory;
import org.apache.log4j.Logger;
import org.apache.xalan.transformer.TrAXFilter;
import org.xml.sax.InputSource;
import org.xml.sax.XMLFilter;
import org.xml.sax.XMLReader;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.ExecutorsHelper;

public class FOPZipMuxSink implements IMuxSink {
	private static final Logger logger = Logger.getLogger(FOPZipMuxSink.class);
	
	protected enum Type {
		File,
		OutputStream;
	};
	
	protected static class DequePayload {
		public IFaucet faucet;

		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof DequePayload))
				return false;
			
			DequePayload other = (DequePayload) obj;
			if (faucet == null)
				return false;
			return faucet.equals(other.faucet);
		}
	}
	
	protected String id;
	protected FopFactory fopFactory;
	protected IFaucet faucetTemplate;
	protected BlockingDeque<DequePayload> deque;
	protected IFaucet faucet;
	protected boolean hookedUp;
	protected String fopTargetMimeType;
	protected Type type;
	protected ZipOutputStream outputStream;
	protected String path;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public FOPZipMuxSink() { }
	
	public FOPZipMuxSink(String id, FopFactory fopFactory, String fopTargetMimeType,
			ZipOutputStream os) {
		this.id = id;
		this.fopFactory = fopFactory;
		this.faucetTemplate = null;
		this.deque = new LinkedBlockingDeque<DequePayload>();
		this.faucet = null;
		this.hookedUp = false;
		this.fopTargetMimeType = fopTargetMimeType;
		this.type = Type.OutputStream;
		this.outputStream = os;
		this.path = null;
	}

	public FOPZipMuxSink(String id, FopFactory fopFactory, String fopTargetMimeType,
			String path) {
		this.id = id;
		this.fopFactory = fopFactory;
		this.faucetTemplate = null;
		this.deque = new LinkedBlockingDeque<DequePayload>();
		this.faucet = null;
		this.hookedUp = false;
		this.fopTargetMimeType = fopTargetMimeType;
		this.type = Type.File;
		this.outputStream = null;
		this.path = path;
	}

	public FOPZipMuxSink(String id, String fopTargetMimeType,
			ZipOutputStream os) {
		this(id, FopFactory.newInstance(), fopTargetMimeType, os);
	}

	public FOPZipMuxSink(String id, String fopTargetMimeType,
			String path) {
		this(id, FopFactory.newInstance(), fopTargetMimeType, path);
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public IFaucet getFaucet() {
		DequePayload payload = deque.peek();
		return payload == null ? null : payload.faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case XML:
			return true;
		case ByteStream:
		case CharStream:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.XML))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the XML content type.");
		if (!(faucet instanceof ISinkTemplate))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must implement the ISinkTemplate interface.");
		if (!((ISinkTemplate) faucet).isClone()) {
			if (faucetTemplate != null && faucetTemplate != faucet)
				throw new XformationException("Hooking a faucet template in an " +
						"already set mux faucet");
			faucetTemplate = faucet;
			return;
		}
		if (hookedUp && faucet == this.faucet) {
			// Effectively a NO-OP.
			return;
		}
		
		this.faucet = faucet;
		this.hookedUp = true;
		
		faucet.hookupSink(this);
	}
	
	@Override
	public void registerSource(Object source) {
		IFaucet faucet = (IFaucet) source;
		
		DequePayload payload = new DequePayload();
		payload.faucet = faucet;

		if (deque.contains(payload))
			return;
		
		if (logger.isTraceEnabled())
			logger.trace("Registering faucet: " + faucet + " with the mux deque");
		
		deque.addLast(payload);
		
	}
	
	@Override
	public String dumpPipeline() {
		DequePayload payload = deque.peekFirst();
		return getClass().getName() + ": " +
			(payload == null ? "null" : payload.faucet);
	}

	@Override
	public void dispose() {
		switch (type) {
		case File:
			try {
				outputStream.close();
			} catch (IOException e) {
				throw new XformationException("Unable to close output stream", e);
			}
			break;
		case OutputStream:
			// Caller owns the output stream. It is their responsibility to close it.
			break;
		}
	}

	private XMLReader getRealReader(Object previousFilter) {
		if (previousFilter instanceof XMLFilter) {
			XMLFilter filter = (XMLFilter) previousFilter;
			return getRealReader(filter.getParent());
		} else {
			return (XMLReader) previousFilter;
		}
	}
	
	@Override
	public void consume() {
		if (!hookedUp && faucetTemplate == null)
			throw new XformationException("Sink has not been set up correctly: " +
					"faucet has not been set");
		
		switch (type) {
		case File:
			try {
				outputStream = new ZipOutputStream(new BufferedOutputStream(
						new FileOutputStream(path)));
			} catch (FileNotFoundException e) {
				throw new XformationException("Unable to create output stream", e);
			}
			break;
		case OutputStream:
			break;
		}
		
		if (!hookedUp) {
			// This is an initial run up the chain. Propagate the consumption of the
			// event. This will start filling up the faucet and input stream deque.
			if (faucetTemplate instanceof IPipelineItem) {
				((IPipelineItem) faucetTemplate).consume(this);
			}
		}
			
		try {
			int counter = 0;
			while (true) {
				DequePayload payload = null;
				try {
					payload = deque.takeFirst();
				} catch (InterruptedException ie) { }
				if (payload == null)
					break;
				
				IFaucet faucet = payload.faucet;
				
				if (logger.isTraceEnabled())
					logger.trace("Retrieved faucet: " + faucet + " from the mux deque");
				
				// If a null has been sent down to the deque then we must exit.
				if (faucet == null)
					break;

				ZipEntry entry = new ZipEntry("entry" + counter++);
				outputStream.putNextEntry(entry);
	
			    InputSource inputSource = null;
			    if (faucet instanceof IPipelineItem) {
			    	inputSource = (InputSource) ((IPipelineItem) faucet).consume(this);
			    }
		
				FOUserAgent foUserAgent = fopFactory.newFOUserAgent();
				
			    // Set up the transformer to process the SAX events generated
			    // by the last filter in the chain
			    try {
					Object faucetObject = faucet.getSource(ContentType.XML);
					if (logger.isTraceEnabled())
						logger.trace("Sink is using reader: " + faucetObject);
					XMLFilter filter = (XMLFilter) faucetObject;
					if (!(filter instanceof TrAXFilter))
						throw new XformationException("Unsupported transformation class " +
								filter.getClass().getName());
				    Transformer transformer = ((TrAXFilter) filter).getTransformer();
					Fop fop = fopFactory.newFop(fopTargetMimeType, foUserAgent, outputStream);
					Result result = new SAXResult(fop.getDefaultHandler());
				    SAXSource transformSource = new SAXSource(getRealReader(filter), inputSource);
				    transformer.transform(transformSource, result);
			    } catch (Exception e) {
			    	throw new XformationException("Unable to set up transform", e);
			    }
				outputStream.closeEntry();
				
				faucet.dispose();
			}
			outputStream.close();
			faucetTemplate.dispose();
		} catch (IOException ioe) {
			logger.error("Error while consuming input", ioe);
			throw new XformationException("Unable to transform stream", ioe);
		}
	}

	@Override
	public Future<?> consumeAsynchronously() {
		return ExecutorsHelper.getInstance().submitInProc(new Runnable() {
			@Override
			public void run() {
				try {
					consume();
				} catch (Throwable t) {
					logger.error("Unable to consume faucet", t);
				}
			}
		});
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		this.fopFactory = FopFactory.newInstance();
		this.faucetTemplate = null;
		this.deque = new LinkedBlockingDeque<DequePayload>();
		this.faucet = null;
		this.hookedUp = false;
		this.fopTargetMimeType = in.readUTF();
		this.type = Type.File;
		this.outputStream = null;
		this.path = in.readUTF();
	}

	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		if (type != Type.File)
			throw new XformationException(
					"Only FOPZipMuxSinks of type Type.File can be streamed.");
		
		out.writeUTF(id);
		out.writeUTF(fopTargetMimeType);
		out.writeUTF(path);
	}

}
