/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.sink;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.Reader;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.concurrent.Future;

import javax.sql.DataSource;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.DataSourceManager;
import com.koutra.dist.proc.util.ExecutorsHelper;

public abstract class ReaderJDBCSink implements ISink {
	private static final Logger logger = Logger.getLogger(StreamSink.class);
	
	protected String id;
	protected IFaucet faucet;
	protected boolean hookedUp;
	protected Reader reader;
	protected DataSource ds;
	protected String dataSourceName;
	protected Connection connection;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public ReaderJDBCSink() { }
	
	public ReaderJDBCSink(String id, DataSource ds) {
		this.id = id;
		this.faucet = null;
		this.hookedUp = false;
		this.reader = null;
		this.ds = ds;
		this.dataSourceName = null;
		this.connection = null;
	}

	public ReaderJDBCSink(String id, String dataSourceName) {
		this.id = id;
		this.faucet = null;
		this.hookedUp = false;
		this.reader = null;
		this.ds = null;
		this.dataSourceName = null;
		this.connection = null;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public IFaucet getFaucet() {
		return faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case CharStream:
			return true;
		case ByteStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.CharStream))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the CharStream content type.");
		if (faucet instanceof IDemuxFaucet)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IDemuxFaucet instance. This is not allowed.");
		if (faucet instanceof IFaucetTemplate)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IFaucetTemplate instance. This is not allowed.");
		if (hookedUp && this.faucet == faucet) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUp)
			throw new XformationException("Trying to hook up an already hooked up sink.");
		
		this.faucet = faucet;
		this.hookedUp = true;
		
		this.faucet.hookupSink(this);
	}
	
	@Override
	public void registerSource(Object source) {
		// Only used in mux/demux situations.
	}
	
	@Override
	public String dumpPipeline() {
		return getClass().getName() + ": " + reader + "->" +
				(dataSourceName == null ? dataSourceName : ds);
	}

	@Override
	public void dispose() {
		try {
			if (connection != null)
				connection.close();
		} catch (SQLException e) {
			throw new XformationException("Unable to close connection");
		}
		
		faucet.dispose();
	}
	
	public abstract void transformBuffer(Connection connection,
			char[] buffer, int off, int len);

	@Override
	public void consume() {
		if (!hookedUp)
			throw new XformationException("Sink has not been set up correctly: " +
					"faucet has not been set");
		
		if (faucet instanceof IPipelineItem) {
			((IPipelineItem) faucet).consume(this);
		}
		
		try {
			if (dataSourceName != null) {
				ds = DataSourceManager.getInstance().getDataSource(dataSourceName);
			}
			connection = ds.getConnection();
			connection.setAutoCommit(true);
			reader = (Reader) faucet.getSource(ContentType.CharStream);
			int count;
			char[] buffer = new char[8 * 1024];
			if (logger.isTraceEnabled())
				logger.trace("Using the reader " + reader + " in the mux sink");
			while ((count = reader.read(buffer)) != -1) {
				if (logger.isTraceEnabled()) {
					String readInput = new String(buffer, 0, count);
					logger.trace("Read " + readInput + " from reader " +
							reader);
				}
				transformBuffer(connection, buffer, 0, count);
			}
		} catch (Exception ioe) {
			throw new XformationException("Unable to transform stream", ioe);
		}
	}

	@Override
	public Future<?> consumeAsynchronously() {
		return ExecutorsHelper.getInstance().submitInProc(new Runnable() {
			@Override
			public void run() {
				try {
					consume();
				} catch (Throwable t) {
					logger.error("Unable to consume faucet", t);
				}
			}
		});
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		this.faucet = null;
		this.hookedUp = false;
		this.reader = null;
		this.ds = null;
		this.dataSourceName = in.readUTF();
		this.connection = null;
	}

	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		if (dataSourceName == null)
			throw new XformationException(
					"Only ResultSetFaucet instances that have the data source name " +
					"specified can be streamed.");
		
		out.writeUTF(id);
		out.writeUTF(dataSourceName);
	}

}
