/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.sink;

import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.concurrent.Future;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.stream.StreamResult;

import org.apache.log4j.Logger;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.ExecutorsHelper;

/**
 * A sink implementation that drives the SAX XML transformation. When using the
 * constructor that supplies the output stream, the caller has the responsibility
 * of closing the output stream when the transformation is completed.
 * 
 * @author Pafsanias Ftakas
 */
public class XMLSink implements ISink {
	private static final Logger logger = Logger.getLogger(XMLSink.class);
	
	protected enum Type {
		File,
		OutputStream;
	};
	
	protected String id;
	protected boolean hookedUp;
	protected IFaucet faucet;
	protected Object faucetObject;
	protected Type type;
	protected OutputStream outputStream;
	protected String path;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public XMLSink() { }
	
	public XMLSink(String id, OutputStream os) {
		this.id = id;
		this.hookedUp = false;
		this.faucet = null;
		this.faucetObject = null;
		this.type = Type.OutputStream;
		this.outputStream = os;
		this.path = null;
	}

	public XMLSink(String id, String path) {
		this.id = id;
		this.hookedUp = false;
		this.faucet = null;
		this.faucetObject = null;
		this.type = Type.File;
		this.outputStream = null;
		this.path = path;
	}

	@Override
	public String getId() {
		return id;
	}
	
	@Override
	public IFaucet getFaucet() {
		return faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case XML:
			return true;
		case ByteStream:
		case CharStream:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.XML))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the XML content type.");
		if (faucet instanceof IDemuxFaucet)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IDemuxFaucet instance. This is not allowed.");
		if (faucet instanceof IFaucetTemplate)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IFaucetTemplate instance. This is not allowed.");
		if (hookedUp && this.faucet == faucet) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUp)
			throw new XformationException("Trying to hook up an already hooked up sink.");
		
		this.faucet = faucet;
		this.hookedUp = true;
		
		this.faucet.hookupSink(this);
	}
	
	@Override
	public void registerSource(Object source) {
		// Only used in mux/demux situations.
	}
	
	@Override
	public String dumpPipeline() {
		return getClass().getName() + ": " + faucetObject + "->" + outputStream;
	}

	@Override
	public void dispose() {
		switch (type) {
		case File:
			try {
				outputStream.close();
			} catch (IOException e) {
				throw new XformationException("Unable to close output stream", e);
			}
			break;
		case OutputStream:
			// Caller owns the output stream. It is their responsibility to close it.
			break;
		}
		
		faucet.dispose();
	}
	
	@Override
	public void consume() {
		if (!hookedUp)
			throw new XformationException("Sink has not been set up correctly: " +
					"faucet has not been set");
		
		switch (type) {
		case File:
			try {
				outputStream = new BufferedOutputStream(new FileOutputStream(path));
			} catch (FileNotFoundException e) {
				throw new XformationException("Unable to create output stream", e);
			}
			break;
		case OutputStream:
			break;
		}

		// Set up the output stream
	    StreamResult result = new StreamResult(outputStream);
	    
	    InputSource input = null;
	    if (faucet instanceof IPipelineItem) {
	    	input = (InputSource) ((IPipelineItem) faucet).consume(this);
	    }

	    // Set up the transformer to process the SAX events generated
	    // by the last filter in the chain
	    try {
	    	faucetObject = faucet.getSource(ContentType.XML);
			if (logger.isTraceEnabled())
				logger.trace("Sink is using reader: " + faucetObject);
		    SAXTransformerFactory stf = (SAXTransformerFactory) TransformerFactory.newInstance();
		    Transformer transformer = stf.newTransformer();
		    SAXSource transformSource = new SAXSource((XMLReader) faucetObject, input);
		    transformer.transform(transformSource, result);
	    } catch (Exception e) {
	    	throw new XformationException("Unable to set up transform", e);
	    }
	}

	@Override
	public Future<?> consumeAsynchronously() {
		return ExecutorsHelper.getInstance().submitInProc(new Runnable() {
			@Override
			public void run() {
				try {
					consume();
				} catch (Throwable t) {
					logger.error("Unable to consume faucet", t);
				}
			}
		});
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		this.hookedUp = false;
		this.faucet = null;
		this.faucetObject = null;
		this.type = Type.File;
		this.outputStream = null;
		this.path = in.readUTF();

	}

	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		if (type != Type.File)
			throw new XformationException(
					"Only XMLSinks of type Type.File can be streamed.");
		
		out.writeUTF(id);
		out.writeUTF(path);
	}

}
