/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.util;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.RejectedExecutionException;

import org.apache.log4j.Logger;
import org.jgroups.ChannelException;
import org.jgroups.util.Streamable;
import org.jgroups.util.Util;

import com.koutra.dist.proc.cluster.ClusterExecutorService;
import com.koutra.dist.proc.cluster.ClusterRunnable;
import com.koutra.dist.proc.cluster.IClusterRunnable;

/**
 * A helper singleton class that provides thread pooling services to the rest
 * of the d-proc system.
 * 
 * @author Pafsanias Ftakas
 */
public class ExecutorsHelper {
	private static final Logger logger = Logger.getLogger(ExecutorsHelper.class);
	
	private static final ExecutorsHelper singleton = new ExecutorsHelper();
	
	public static ExecutorsHelper getInstance() {
		return singleton;
	}
	
	private ExecutorService executorService;
	private ClusterExecutorService clusterExecutorService;
	
	private ExecutorsHelper() {
		executorService = Executors.newCachedThreadPool();
		clusterExecutorService = null;
	}
	
	private synchronized void assureClusterExecutorServiceExists() {
		try {
			if (clusterExecutorService == null)
				clusterExecutorService = new ClusterExecutorService();
		} catch (ChannelException e) {
			logger.info(
					"Unable to create cluster executor, will revert to JVM executors", e);
		}
	}
	
	public void executeInProc(Runnable command) {
		executorService.execute(command);
	}
	
	public Future<?> submitInProc(Runnable command) {
		return executorService.submit(command);
	}
	
	public synchronized void shutdown() {
		executorService.shutdown();
		
		if (clusterExecutorService != null)
			clusterExecutorService.shutdown();
	}
	
	public void executeInCluster(Runnable command) {
		assureClusterExecutorServiceExists();
		
		if (!(command instanceof IClusterRunnable))
			throw new IllegalArgumentException(
					"Argument to the ClusterExecutor class must be implementing the " +
					"ClusterRunnable interface. Class is: " +
					command.getClass().getCanonicalName());
		
		ClusterRunnable actualCommand = (ClusterRunnable) command;
		Streamable payload = actualCommand.getWirePayload();
		if (clusterExecutorService == null) {
			try {
				Object boundaryTemplate = Util.objectFromByteBuffer(
						Util.objectToByteBuffer(payload));
				actualCommand.setConainingObject(boundaryTemplate);
				executorService.execute(actualCommand);
			} catch (Exception e) {
				logger.error("Unable to mock execute in cluster", e);
			}
		} else {
			clusterExecutorService.execute(actualCommand);
		}
	}
	
	public Future<?> submitInCluster(Runnable command) {
		assureClusterExecutorServiceExists();
		
		if (!(command instanceof IClusterRunnable))
			throw new IllegalArgumentException(
					"Argument to the ClusterExecutor class must be implementing the " +
					"ClusterRunnable interface. Class is: " +
					command.getClass().getCanonicalName());
		
		ClusterRunnable actualCommand = (ClusterRunnable) command;
		Streamable payload = actualCommand.getWirePayload();
		if (clusterExecutorService == null) {
			try {
				Object boundaryTemplate = Util.objectFromByteBuffer(
						Util.objectToByteBuffer(payload));
				actualCommand.setConainingObject(boundaryTemplate);
				return executorService.submit(actualCommand);
			} catch (Exception e) {
				logger.error("Unable to mock execute in cluster", e);
				throw new RejectedExecutionException("Unable to mock execute in cluster", e);
			}
		} else {
			return clusterExecutorService.submit(actualCommand);
		}
	}
}
