/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.xml;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Iterator;

import org.apache.log4j.Logger;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;
import org.xml.sax.helpers.XMLFilterImpl;

import com.koutra.dist.proc.pipeline.XMLFilterDemuxPipelineItem;

public class XPathXMLMuxFilter extends XMLFilterImpl implements XMLMuxFilter {
	private static final Logger logger = Logger.getLogger(XPathXMLMuxFilter.class);
	
	protected XMLFilterDemuxPipelineItem demux;
	protected TraversedPath path;
	protected String xPath;
	protected int fragmentNumber;
	protected boolean rootEncounter;
	
	/**
	 * @deprecated use the initializing constructor instead.
	 */
	public XPathXMLMuxFilter() { }
	
	public XPathXMLMuxFilter(String xPath) {
		this.demux = null;
		this.path = new TraversedPath();
		this.xPath = xPath;
		this.fragmentNumber = 1;
		this.rootEncounter = false;
	}
	
	@Override
	public void setDemuxFilter(XMLFilterDemuxPipelineItem demux) {
		this.demux = demux;
	}
	
	protected void endAllAncestors() throws SAXException {
		Iterator<SAXElement> iter = path.getElementsReverseIterator();
		while (iter.hasNext()) {
			SAXElement se = (SAXElement) iter.next();
			super.endElement(se.getUri(), se.getLocalName(), se.getQName());
		}
	}
	
	protected void updateRootAncestor() {
		SAXElement rootElement = path.getElement(0);
		if (rootElement.getAttributes().getIndex("fragment") != -1) {
			rootElement.getAttributes().removeAttribute(
					rootElement.getAttributes().getIndex("fragment"));
		}
		rootElement.getAttributes().addAttribute("", "fragment", "fragment", "CDATA", 
				String.valueOf(fragmentNumber));
	}
	
	protected void restoreAllAncestors() throws SAXException {
		Iterator<SAXElement> iter = path.getElementsIterator();
		while (iter.hasNext()) {
			SAXElement se = iter.next();
			
			AttributesImpl attributes = new AttributesImpl(se.getAttributes());
			attributes.addAttribute("", "continued", "continued", "CDATA", "true");
			super.startElement(se.getUri(), 
					se.getLocalName(), se.getQName(), attributes);
		}
	}
	
	protected void breakXML() throws SAXException {
		if (logger.isTraceEnabled())
			logger.trace("About to break the XML document at: " + path.toString());
		
		endAllAncestors();
		
		super.endDocument();
		
		fragmentNumber++;
		
		demux.switchPipeline();
		
		super.startDocument();
		
		updateRootAncestor();
		
		restoreAllAncestors();
	}
	
	@Override
	public void startDocument() throws SAXException {
		logger.trace("Starting a document");
		super.startDocument();
		path.startDocument();
	}
	
	@Override
	public void endDocument() throws SAXException {
		path.endDocument();
		super.endDocument();
		logger.trace("End of document reached");
		
		demux.signalTermination();
	}
	
	@Override
	public void startElement(String uri, String localName, String name,
			Attributes atts) throws SAXException {
		if (logger.isTraceEnabled())
			logger.trace("Element (" + SAXElement.getName(localName, name) +
					") starting at: " + path.toString());
		
		Attributes potentiallyModifiedAtts = atts;
		if (!rootEncounter) {
			AttributesImpl modifiedAtts = new AttributesImpl(atts);
			if (modifiedAtts.getIndex("fragment") != -1)
				modifiedAtts.removeAttribute(modifiedAtts.getIndex("fragment"));
			modifiedAtts.addAttribute("", "fragment", "fragment", "CDATA", 
					String.valueOf(fragmentNumber));
			potentiallyModifiedAtts = modifiedAtts;
			rootEncounter = true;
		}
		
		super.startElement(uri, localName, name, potentiallyModifiedAtts);
		path.startElement(uri, localName, name, atts);
	}
	
	@Override
	public void endElement(String uri, String localName, String name)
			throws SAXException {
		boolean needToBreak = path.compareTo(xPath, -1);

		path.endElement(uri, localName, name);
		super.endElement(uri, localName, name);
		if (logger.isTraceEnabled())
			logger.trace("Element (" + SAXElement.getName(localName, name) +
					") ending at: " + path.toString());
		
		if (needToBreak)
			breakXML();
	}
	
	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.demux = null;
		this.path = new TraversedPath();
		this.xPath = in.readUTF();
		this.fragmentNumber = 1;
		this.rootEncounter = false;
	}
	
	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		out.writeUTF(xPath);
	}
	
}
